/**********************************************************************
 *
 * dbActivity.c
 * version 1.2
 *
 * dbActivity is a command line tool that allows you to display the
 * user activity that is logged in a Notes database. You can optionally
 * filter the activity by user name and/or time periods, which is useful
 * for determining whether or not certain users actually accessed a
 * database on a given set of days.
 *
 * To compile this program, you need to have your compiler all set up
 * to use the Notes C-API libraries (I think I used the 5.06 libraries,
 * but other versions will probably work too, as long as they're not
 * too old). The final program executable requires a Notes client to
 * be installed on the computer, and you will normally have to run
 * the program from the Notes program directory. I have only tested this
 * compiling on Windows...other operating systems may require some tweaks
 * to the code.
 *
 * This is the first open source version of dbActivity. Please don't
 * laugh at the code. It's kind of hacked together from lots of places,
 * and if I had the time to write it all over again I would. I'd also
 * like to have commented everything a little more, but I don't really
 * have time for that right now either.
 *
 * A lot of the main code (in the PrintActivity function) is taken
 * primarily from the Notes C-API help files for the NSFDbGetUserActivity
 * function. It seemed silly to rewrite all that code, because it worked
 * pretty well; however, the inclusion of code that someone else wrote
 * added to the randomness that you'll see in the coding style below.
 * Also, I'm really not much of a C programmer, so the more experienced
 * readers will probably cringe at the way I did some of this.
 * Oh well...somehow it all seems to work.
 *
 * I also made use of a few generic pieces of code that I've written,
 * which probably seem a bit mysterious at first glance. They are:
 *
 * => GetDbList - a set of functions I wrote that allow you to easily
 *                step through a list of databases on a Notes server
 *
 * => wildcard - a function that allows you to see if a given string
 *               matches a wildcard specification (using ? and/or *)
 *
 * => StderrToOut - a function that swaps the output of stderr and stdout.
 *                  This is a pretty funky thing to want to do, but in
 *                  Notes API command line programs you can run into a
 *                  problem if you're piping stdout to a text file and
 *                  the API prompts you for a password for your Notes ID.
 *                  You can't really intercept the password prompt, and
 *                  on some versions of Notes it will prompt you using
 *                  stdout and stdin. If you've swapped stderr and stdout,
 *                  then you're okay even if you're piping your output,
 *                  because the password prompt will show up in the console
 *                  instead of in the piped file. However, all your printf
 *                  text will also show up on the console, so I also have
 *                  a function called printerr that works just like printf
 *                  only it sends text to stderr (which is now stdout)
 *                  instead of stdout (which is now stderr). Confusing, huh?
 *
 * All of these files should be zipped together in any distribution of this
 * code. At the time of this writing, all the current versions of the code 
 * is still available at the http://www.nsftools.com site (in the "Tools"
 * section), although my goal is to turn this into an OpenNTF project so
 * that other people can contribute to the codestream. Check there too:
 * http://www.OpenNTF.org (or http://www.notesoss.org for the oldtimers).
 *
 * Written by Julian Robichaux
 * http://www.nsftools.com
 *
 * You may use this code in any way you'd like, as long as you don't
 * hold me liable for anything, and you don't tell people that you 
 * wrote it yourself. If you're feeling really generous, please include
 * my name and website whenever you use this code in your programs.
 * Thanks!
 *
 * HISTORY
 * version 1.1   - first open source version of the program
 * version 1.2   - added functionality to allow the user to specify
 *                 names to exclude using the /x parameter
 **********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <global.h>
#include <nsfdb.h>
#include <misc.h>
#include <ostime.h>
#include <osfile.h>
#include <osmem.h>
#include <osmisc.h>

#include "StderrToOut.h"
#include "GetDbList.h"
#include "wildcard.h"

// most of the globals will be set as command line parameters
// in the ProcessArgs function
BOOL	PrintSummaryInfo = FALSE;
char	dbFilter[MAXPATH] = "*";
char	nameFilter[MAXPATH] = "*";
char	nameExcludeFilter[MAXPATH] = "";
int		numberOfDays = 0;
int		startDays = -1;
int		endDays = -1;
char	server_name[MAXPATH];
char	delim = '\t';

// forward declarations, since I was too lazy to create a
// proper header file for all this
STATUS GetActivity (char *serverName, char *dbName);
STATUS PrintActivity (DBHANDLE db_handle);
void PrintError (STATUS nError);
int ProcessArgs (int argc, char *argv[]);
void PrintUsage ();


/*********************************************************************************
 * NotesMain is the primary entry point for a Notes API command line
 * program, so it's the primary entry point for us too (see the API
 * documentation if you want to confirm this; otherwise, you'll just
 * have to trust me). This is where everything gets kicked off.
 *********************************************************************************/
STATUS LNPUBLIC NotesMain (int argc, char *argv[])
{
	STATUS nError;
	char dbName[MAXPATH];
	char dbNameOnly[MAXPATH];

	// process the command line arguments (if we returned FALSE,
	// something wasn't right so we should just exit)
	if (!ProcessArgs(argc, argv))
		return (NOERROR);

	// swap the stdout and stderr streams, in case the user is piping output
	// to a text file (that way, if Notes prompts us for a password, we'll
	// see it on the console instead of in the text file, which does us no
	// good). In order to print anything to the piped text file (or to the
	// console screen, if no piping is taking place), use the "printerr"
	// function -- which is why you'll see calls to printerr all over the
	// place in the code that follows
	OutToErr();

	// we're allowing wildcard specifications for database names, so we can
	// use the GetDbList functions to step through all the databases on a
	// given server and figure out which ones match our filter
 	while (GetNextDbName(server_name, "", dbName, &nError))
 	{
 		GetDbNameOnly(dbNameOnly);
 		if ((IsWildcardMatch(dbFilter, dbName, FALSE)) || 
			(IsWildcardMatch(dbFilter, dbNameOnly, FALSE)))
		{
			// I know that separators like this are cheesy, but they
			// help me see everything better when I've got multiple
			// matches...
			printerr("\n===========================\n");
 			GetActivity(server_name, dbName);
		}
 	}

	return(NOERROR);

}


/*********************************************************************************
 * GetActivity is the function that tries to open a database and send it
 * to the PrintActivity function. It's probably silly to have this as a
 * separate function like this instead of wrapping it all up in the
 * PrintActivity function itself, but that's just how I wrote it.
 *********************************************************************************/
STATUS GetActivity (char *serverName, char *dbName)
{
	STATUS nError;
	char dbPath[MAXPATH];
	DBHANDLE hDb;

	/* try to get the database path */
	if (nError = OSPathNetConstruct(NULL, serverName, dbName, dbPath))
	{
		printerr("Cannot create path to database %s\n", dbName);
		return nError;
	}

	/* try to open the database */
	nError = NSFDbOpen(dbPath, &hDb);
	if (hDb)
	{
		printerr("Activity for database: %s\n\n", dbName);
		nError = PrintActivity(hDb);
		if (nError)
		{
			/* if the PrintActivity routine returned an error, then either user activity
			    isn't being logged (error 578) or something else funny is going on,
				probably an access issue. In any case, error out with a message to 
				the user. */
			if (ERR(nError == 578))
			{
				printerr("User activity is not currently being logged in this database.\n");
				printerr("Go to Database Properties - User Activity, and check 'Record Activity'\n");
			}  else  {
				PrintError(nError);
			}
		}
		NSFDbClose(hDb);

	}  else  {
		printerr("Unable to open database: %s\n\n", dbName);
		PrintError(nError);
	}

	return nError;
}


/*********************************************************************************
 * Okay, the PrintActivity function is the meat of the program. It goes
 * through the user activity logs (if there are any) and sends the
 * information to the console. It also takes into account a few of our
 * global variables (yeah, I know that's not a very generic implementation
 * of this function, but whatever), for things like filtering the user names
 * and dates, and formatting the output text. This code is a modification of
 * the NSFDbGetUserActivity sample code from the C-API help files.
 *********************************************************************************/
STATUS PrintActivity (DBHANDLE db_handle)
{
	/* Get the database activity summary. The database has already been opened. */
	/* From the C API help files for NSFDbGetUserActivity (with modifications) */
	STATUS nError = NOERROR;
	TIMEDATE timeNow;
	TIMEDATE startTime, endTime, activityTime;
	DBACTIVITY DbSummaryActivity;
	DBACTIVITY_ENTRY *pDbUserActivity, DbUserActivity;
	HANDLE hDbUserActivity;
	WORD wUserCount, wLen;
	char *szUserName;
	char szTimedate[MAXALPHATIMEDATE+1];
	LONG lRptPeriod, timeDiff;
	WORD i;

	/* Try to get a handle to the User Activity log in the database */
	nError = NSFDbGetUserActivity (db_handle, 0L, &DbSummaryActivity, &hDbUserActivity, &wUserCount);
	if (nError)
	{
		printerr("Unable to get User Activity Log\n");
		return (nError);
	}

	/* Get the date/time stuff we'll be using */
	OSCurrentTIMEDATE(&timeNow);
	TimeExtractLocal(&timeNow, FALSE, &startTime);
	TimeDateAdjust(&startTime, 0, 0, 0, (0-startDays), 0, 0);
	TimeExtractLocal(&timeNow, FALSE, &endTime);
	TimeDateAdjust(&endTime, 0, 0, 0, (0-endDays), 0, 0);

	/* Obtain a pointer to the first structure in an array of
	   DBACTIVITY_ENTRY structures. The user names follow the array */
	pDbUserActivity = OSLock (DBACTIVITY_ENTRY, hDbUserActivity);
	if (delim == ',')
		printerr("Date/Time,Reads,Writes,User Name\n");
	else
		printerr("Date/Time\t\tReads\tWrites\tUser Name\n");

	for (i = 0; i < wUserCount; i++)
	{
		DbUserActivity = pDbUserActivity[i];
		nError = ConvertTIMEDATEToText (NULL, NULL, &(DbUserActivity.Time),
						szTimedate, MAXALPHATIMEDATE, &wLen);
		if (nError)
		{
			printerr("Error converting time/date stamp on record %d\n", i);
			continue;
		}

		szTimedate[wLen] = '\0';
		szUserName = NSFDbGetActivityUserNamePtr (pDbUserActivity, i);

		/* figure out if this record matches our time criteria */
		timeDiff = 1;
		TimeExtractLocal(&(DbUserActivity.Time), FALSE, &activityTime);

		if (endDays >= 0)
		{
			// both of these methods work, but I suspect the second is faster:
			//timeDiff = TimeDateDifference(&check_end_td, &tdComp);
			timeDiff = TimeExtractJulianDate(&endTime) - TimeExtractJulianDate(&activityTime);
		}

		if ((startDays >= 0) && (timeDiff >= 0))
		{
			//timeDiff = TimeDateDifference(&tdComp, &check_begin_td);
			timeDiff = TimeExtractJulianDate(&activityTime) - TimeExtractJulianDate(&startTime);
		}

		/* if the user name matches our wildcard filter, print out DBACTIVITY_ENTRY Info */
		if ( (timeDiff >= 0) && (IsWildcardMatch(nameFilter, szUserName, FALSE)) && 
			 (!IsWildcardMatch(nameExcludeFilter, szUserName, FALSE)) )
		{
			printerr("%s%c%d%c%d%c%s\n", szTimedate, delim, DbUserActivity.Reads, delim,
						DbUserActivity.Writes, delim, szUserName);
		}

	}

	OSUnlock (hDbUserActivity);
	OSMemFree (hDbUserActivity);

	/* We're done. Now print the DBACTIVITY summary info */
	if (PrintSummaryInfo)
	{
		printerr("\n\n");
		printerr("Database Summary Info (all users):\n");
		printerr("Summary Period%cReads%cWrites%cUses\n", delim, delim, delim);
		printerr("Last Day%c%d%c%d%c%d\n", delim,
			DbSummaryActivity.PrevDayReads, delim,
			DbSummaryActivity.PrevDayWrites, delim,
			DbSummaryActivity.PrevDayUses);
		printerr("Last Week%c%d%c%d%c%d\n", delim,
			DbSummaryActivity.PrevWeekReads, delim,
			DbSummaryActivity.PrevWeekWrites, delim,
			DbSummaryActivity.PrevWeekUses);
		printerr("Last Month%c%d%c%d%c%d\n", delim,
			DbSummaryActivity.PrevMonthReads, delim,
			DbSummaryActivity.PrevMonthWrites, delim,
			DbSummaryActivity.PrevMonthUses);

		lRptPeriod = TimeDateDifference(&(DbSummaryActivity.Last), &(DbSummaryActivity.First));
		/* Convert from seconds to days */
		if (lRptPeriod % (60*60*24))
			lRptPeriod = lRptPeriod / (60*60*24) + 1;
		else
			lRptPeriod = lRptPeriod / (60*60*24);

		printerr("Last %lu days%c%d%c%d%c%d\n", lRptPeriod, delim,
			DbSummaryActivity.Reads, delim,
			DbSummaryActivity.Writes, delim,
			DbSummaryActivity.Uses);
	}

	return (nError);

}


/*********************************************************************************
 * PrintError is just a convenience function to send a STATUS error to
 * the console (using printerr in this case) as a nicely formatted
 * message.
 *********************************************************************************/
void PrintError (STATUS nError)
{
	#define ERR_TEXT 256
	char errorString[ERR_TEXT];

	if (OSLoadString(NULLHANDLE, ERR(nError), errorString, ERR_TEXT-1))
		printerr("(Error %d: %s)\n", ERR(nError), errorString);
}


/*********************************************************************************
 * ProcessArgs is what we use to process the command line arguments and
 * set our global variables. If the minimum arguments have not been
 * specified, the help message for this program will be displayed and
 * this function will return FALSE. Otherwise, we'll return TRUE.
 *********************************************************************************/
int ProcessArgs (int argc, char *argv[])
{
	int i;
	char *token;
	int stringCount = 0;
	char *argString[3];

	// if there were no command line arguments at all, just
	// print the help and return
	if (argc < 2)
	{
		PrintUsage();
		return FALSE;
	}

	// otherwise, we accept either "-" or "/" as valid switch prefixes;
	// anything without a switch prefix is treated as either a server
	// name or a database name (depending on how many non-switched args
	// we end up getting)
	for (i = 1; i < argc; i++)
	{
		if ((argv[i][0] == '-') | (argv[i][0] == '/'))
		{
			switch (argv[i][1])
			{
			case '?' :
				// if any of the args are a question mark,
				// just print the help and leave
				PrintUsage();
				return FALSE;

			case 'c' :
			case 'C' :
				// "c" is the switch for comma-delimited output
				delim = ',';
				break;

			case 'i' :
			case 'I' :
				// "i" is the switch for printing summary info after
				// we've printed the desired activity records
				PrintSummaryInfo = TRUE;
				break;

			case 'f' :
			case 'F' :
				// "f:" is the switch for a wildcard filter for user
				// names (the filter comes after the ":")
				token = strtok( argv[i], ":" );
				token = strtok( NULL, "\0" );
				if (token != NULL)
					strcpy(nameFilter, token);
				break;

			case 'x' :
			case 'X' :
				// "x:" is the switch for a wildcard filter for user
				// names to exclude (the filter comes after the ":")
				token = strtok( argv[i], ":" );
				token = strtok( NULL, "\0" );
				if (token != NULL)
					strcpy(nameExcludeFilter, token);
				break;

			case 's' :
			case 'S' :
				// "s:" is the switch for how many days ago the start
				// date that we're looking for is (comes after the ":")
				token = strtok( argv[i], ":" );
				token = strtok( NULL, "\0" );
				if (token != NULL)
					startDays = atoi(token);
				break;

			case 'e' :
			case 'E' :
				// "e:" is the switch for how many days ago the end
				// date that we're looking for is (comes after the ":")
				token = strtok( argv[i], ":" );
				token = strtok( NULL, "\0" );
				if (token != NULL)
					endDays = atoi(token);
				break;

			default :
				// if we don't recognize the switch, we'll ignore it
				printerr("Unrecognized command line option:%s (ignored)", argv[i]);
				break;
			}

		}  else  {
			// if this wasn't a switched argument, add it to our list of strings
			// to process when we're done
			if (stringCount <= 1)
			{
				argString[stringCount] = argv[i];
				stringCount++;
			}
		}

	}

	// now that we're done parsing the args, decide what the non-switched
	// strings represented
	switch (stringCount)
	{
	case 0 :
		printerr("** Not enough arguments! **\n");
		PrintUsage();
		return FALSE;
	case 1 :
		strcpy(server_name, "");
		strcpy(dbFilter, argString[0]);
		break;
	default :
		strcpy(server_name, argString[0]);
		strcpy(dbFilter, argString[1]);
		break;
	}

	return TRUE;

}


/*********************************************************************************
 * PrintUsage just displays information about how to use this program
 *********************************************************************************/
void PrintUsage ()
{
	printf("Displays user activity information in a database, if available.\n\n");

	printf("dbActivity [ServerName] <DbName> [/c] [/i] [/f:filter] [/s:days] [/e:days]\n\n");

	printf("  ServerName      [optional] The name of the server that the database is on.\n");
	printf("  DbName          The name of the database to examine (wildcards are allowed).\n");
	printf("  /c              Display information in comma-delimited format.\n");
	printf("  /i              Display summary information at the end.\n");
	printf("  /f:filter       Filter the user names that are retrieved based on a wildcard\n");
	printf("                  specification using * or ? (like *John*).\n");
	printf("  /x:filter       Exclude names that match this wildcard pattern (like *SRV*)\n");
	printf("  /s:days         Get only the entries that are this many days old or newer.\n");
	printf("                  If this option is not specified, there is no start date.\n");
	printf("  /e:days         Get only the entries that are this many days old or older.\n");
	printf("                  If this option is not specified, there is no end date.\n");

	printf("\nIf the server or database name contains spaces, enclose it in quotation marks.\n");
	printf("You can use a combination of the /s and /e options to specify a date range.\n");
	printf("For example, /s:14 /e:7 gets entries from the last 7 to 14 days, /s:1 /e:1\n");
	printf("gets entries from yesterday, and /s:0 /e:0 gets entries from today.\n");

	printf("\ncopyright 2003, Julian Robichaux -- nsftools.com\nversion 1.2\n");
}


